import aiohttp
import asyncio
import functools
import time
from .token_manager import TokenManager
from .ratelimiter import RateLimiter
from .exceptions import TreillageHTTPException, TreillageRateLimitException


def renew_access_token(func):
    @functools.wraps(func)
    async def wrapped(self, *args, **kwargs):
        # Refresh the token 90 seconds before it expires
        if time.time() > self.token_manager.access_token_expiry - 90:
            await self.token_manager.refresh_access_token()
        return await func(self, *args, **kwargs)

    return wrapped


def rate_limit(func):
    @functools.wraps(func)
    async def wrapped(self, *args, **kwargs):
        if self.rate_limiter:
            await self.rate_limiter.get_token()
        return await func(self, *args, **kwargs)

    return wrapped


def retry_on_rate_limit(func):
    @functools.wraps(func)
    async def wrapped(*args, **kwargs):
        while True:
            try:
                return await func(*args, **kwargs)
            except TreillageRateLimitException:
                pass
    return wrapped


class ConnectionManager:
    def __init__(self,
                 base_url: str,
                 credentials,
                 max_connections: int = None,
                 rate_limit_token_regen_rate: int = None
                 ):
        self.__base_url = base_url
        self.__credentials = credentials
        if max_connections is not None:
            self.__connector = aiohttp.TCPConnector(
                limit_per_host=max_connections
            )
        else:
            self.__connector = None
        self.__session = None
        self.__auth_tokens = None
        if rate_limit_token_regen_rate is not None:
            self.__rate_limiter = RateLimiter(
                token_rate=rate_limit_token_regen_rate
            )
        else:
            self.__rate_limiter = None

    @classmethod
    async def create(cls,
                     base_url: str,
                     credentials,
                     max_connections: int = None,
                     rate_limit_token_regen_rate: int = None
                     ):

        self = ConnectionManager(
            base_url,
            credentials,
            max_connections,
            rate_limit_token_regen_rate
        )
        self.__auth_tokens = await TokenManager.create(credentials, base_url)
        if self.connector:
            self.__session = aiohttp.ClientSession(
                timeout=aiohttp.ClientTimeout(total=90),
                connector=self.connector
            )
        else:
            self.__session = aiohttp.ClientSession(
                timeout=aiohttp.ClientTimeout(total=90)
            )
        return self

    async def close(self):
        if self.__session is not None:
            await self.__session.close()
            # Sleep to give connections time to close
            await asyncio.sleep(0.250)

    @property
    def token_manager(self) -> TokenManager:
        return self.__auth_tokens

    @property
    def rate_limiter(self) -> RateLimiter:
        return self.__rate_limiter

    @property
    def connector(self) -> aiohttp.TCPConnector:
        return self.__connector

    async def __handle_response(self, response, http_success_code: int = 200):
        if response.status == http_success_code:
            if self.__rate_limiter is not None:
                self.__rate_limiter.last_try_success(True)
            return await response.json()
        else:
            msg = await response.text()
            if response.status == 429:
                if self.__rate_limiter is not None:
                    self.__rate_limiter.last_try_success(False)
                raise TreillageRateLimitException(url=response.url, msg=msg)
            else:
                raise TreillageHTTPException(
                    code=response.status,
                    url=response.url,
                    msg=msg
                )

    def __setup_headers(self, headers: dict = None) -> dict:
        if not headers:
            headers = dict()
        headers["x-fv-sessionid"] = self.__auth_tokens.refresh_token
        headers["Authorization"] = f"Bearer {self.__auth_tokens.access_token}"
        return headers

    @renew_access_token
    @rate_limit
    async def get(
            self,
            endpoint: str,
            params: dict = None,
            headers: dict = None
    ):
        async with self.__session.get(
                url=self.__base_url + endpoint,
                params=params,
                headers=self.__setup_headers(headers)
        ) as response:
            return await self.__handle_response(response, 200)

    @renew_access_token
    @rate_limit
    async def patch(self, endpoint: str, body: dict, headers: dict = None):
        async with self.__session.patch(
                url=self.__base_url + endpoint,
                json=body,
                headers=self.__setup_headers(headers)
        ) as response:
            return await self.__handle_response(response, 200)

    @renew_access_token
    @rate_limit
    async def post(self, endpoint: str, body: dict, headers: dict = None):
        async with self.__session.post(
                url=self.__base_url + endpoint,
                json=body,
                headers=self.__setup_headers(headers)
        ) as response:
            return await self.__handle_response(response, 200)

    @renew_access_token
    @rate_limit
    async def put(self, endpoint: str, body: dict, headers: dict = None):
        async with self.__session.put(
                url=self.__base_url + endpoint,
                json=body,
                headers=self.__setup_headers(headers)
        ) as response:
            return await self.__handle_response(response, 200)
            
    @renew_access_token
    @rate_limit
    async def delete(self, endpoint: str, headers: dict = None):
        async with self.__session.delete(
                url=self.__base_url + endpoint,
                headers=self.__setup_headers(headers)
        ) as response:
            return await self.__handle_response(response, 204)



from typing import List
from .. import ConnectionManager
from .. import TreillageTypeError, TreillageValueError
from .list_paginator import list_paginator


# * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
#                              Contacts
# * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


async def get_contact(
        connection: ConnectionManager,
        contact_id: str,
        fields: List[str] = None
):
    endpoint = f'/core/contacts/{contact_id}'
    params = dict()
    if fields:
        requested_fields = ','.join(*[fields])
        params['requestedFields'] = requested_fields

    return await connection.get(endpoint, params)


async def get_contact_list(connection: ConnectionManager,
                           fields: List[str] = None,
                           first_name: str = None,
                           last_name: str = None,
                           full_name: str = None,
                           nick_name: str = None,
                           person_type: str = None,
                           phone: str = None,
                           email: str = None
                           ):
    endpoint = '/core/contacts'
    params = dict()
    if fields:
        requested_fields = ','.join(*[fields])
        params['requestedFields'] = requested_fields
    if first_name:
        params['firstName'] = first_name
    if last_name:
        params['lastName'] = last_name
    if full_name:
        params['fullName'] = full_name
    if nick_name:
        params['nickName'] = nick_name
    if person_type:
        params['personType'] = person_type
    if phone:
        params['phone'] = phone
    if email:
        params['email'] = email

    async for contact in list_paginator(connection, endpoint, params):
        yield contact


async def create_contact():
    raise NotImplementedError


async def update_contact(connection: ConnectionManager,
                         contact_id: str,
                         person_types: List[str],
                         person_id: dict = None,
                         first_name: str = None,
                         middle_name: str = None,
                         last_name: str = None,
                         is_single_name: bool = None,
                         full_name: str = None,
                         ssn: str = None,
                         birthdate: str = None,
                         notes: str = None,
                         from_company: str = None,
                         specialty: str = None,
                         gender: str = None,
                         language: str = None,
                         marital_status: str = None,
                         is_texting_permitted: bool = None,
                         remarket: bool = None,
                         abbreviated_name: str = None,
                         driver_license_number: str = None,
                         salutation: str = None,
                         bar_number: str = None,
                         fiduciary: str = None,
                         is_minor: bool = None,
                         phones: List = None,
                         emails: List = None,
                         addresses: List = None,
                         ):
    def validate_person_type(person_type) -> bool:
        valid_types = {'Adjuster', 'Attorney', 'Client', 'Court', 'Defendant',
                       'Plaintiff', 'Expert', 'Firm', 'Insurance Company',
                       'Involved Party', 'Judge', 'Medical Provider'}
        if person_type in valid_types:
            return True
        else:
            raise TreillageValueError(
                f"{person_type} not in allowed types: " +
                f"{', '.join(*valid_types)}"
            )

    def validate_marital_status(status) -> bool:
        valid_statuses = {"s", "m", "d", "u", "w"}
        if status in valid_statuses:
            return True
        else:
            raise TreillageValueError(
                f"{status} not in allowed marital statuses of:" +
                f"{', '.join(*valid_statuses)}"
            )

    endpoint = f'/core/contacts/{contact_id}'
    body = dict()
    if person_types:
        body["personTypes"] = []
        for p_type in person_types:
            if validate_person_type(p_type):
                body["personTypes"].append(p_type)
    if person_id and isinstance(person_id, dict):
        body["personId"] = person_id
    if first_name:
        body['firstName'] = first_name
    if middle_name:
        body['middleName'] = middle_name
    if last_name:
        body['lastName'] = last_name
    if is_single_name:
        if isinstance(is_single_name, bool):
            body['isSingleName'] = is_single_name
        else:
            raise TreillageTypeError
    if full_name:
        body['fullName'] = full_name
    if ssn:
        body['ssn'] = ssn
    if birthdate:
        body['birthDate'] = birthdate
    if notes:
        body['notes'] = notes
    if from_company:
        body['fromCompany'] = from_company
    if specialty:
        body["specialty"] = specialty
    if gender:
        body['gender'] = gender
    if language:
        body['language'] = language
    # Single Character Values Only Accepted Values: "s","m","d","u","w"
    if marital_status and validate_marital_status(marital_status):
        body['maritalStatus'] = marital_status
    if is_texting_permitted:
        if isinstance(is_texting_permitted, bool):
            body['isTextingPermitted'] = is_texting_permitted
        else:
            raise TreillageTypeError
    if remarket:
        if isinstance(remarket, bool):
            body['remarket'] = remarket
        else:
            raise TreillageTypeError
    if abbreviated_name:
        body['abbreviatedName'] = abbreviated_name
    if driver_license_number:
        body['driverLicenseNumber'] = driver_license_number
    if salutation:
        body['salutation'] = salutation
    if bar_number:
        body['barNumber'] = bar_number
    if fiduciary:
        body['fiduciary'] = fiduciary
    if is_minor:
        body['isMinor'] = is_minor
    if phones:
        if isinstance(phones, List):
            body['phones'] = phones
        else:
            raise TreillageTypeError
    if emails:
        if isinstance(emails, List):
            body['emails'] = emails
        else:
            raise TreillageTypeError
    if addresses:
        if isinstance(addresses, List):
            body['addresses'] = addresses
        else:
            raise TreillageTypeError

    return await connection.patch(endpoint, body)



class TreillageException(Exception):
    def __init__(self, msg=None, url=None):
        if not msg:
            msg = "An error occurred when trying to access the Filevine API"
            if url is not None:
                msg += f" at {url}"
        super(TreillageException, self).__init__(msg)
        self.url = url
        self.msg = msg


class TreillageHTTPException(TreillageException):
    def __init__(self, code, url=None, msg=None):
        self.code = code
        if not msg:
            msg = f"Received non-2xx HTTP Status Code {code}"
        super(TreillageHTTPException, self).__init__(msg=msg, url=url, )


class TreillageRateLimitException(TreillageHTTPException):
    def __init__(self, url=None, msg=None):
        self.code = 429
        if not msg:
            msg = "Server Rate Limit Exceeded"
        super(TreillageRateLimitException, self).__init__(self.code, url, msg)


class TreillageTypeError(TreillageException, TypeError):
    def __init__(self, expected_type, received_type):
        msg = (f"Type {expected_type} required in API definition, " +
               f"but received {received_type}")
        super(TreillageTypeError, self).__init__(msg=msg)


class TreillageValueError(TreillageException, ValueError):
    def __init__(self, msg=None):
        if not msg:
            msg = "Given parameter does not meet argument requirements"
        super(TreillageValueError, self).__init__(msg=msg)



from setuptools import setup, find_packages

import versioneer

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='treillage',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Unofficial Wrapper library for the Filevine API',
    license='MIT',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/W1ndst0rm/Treillage",
    author='Levi Jumonville',
    packages=find_packages(),
    install_requires=[
        'aiohttp>=3,<4',
        'pandas>=1,<2',
        'PyYAML>=5,<6',
        'PyJWT>=1.6.4,<2',
        'cryptography>=35,<36'
    ],
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Legal Industry",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Internet"
    ],
    python_requires='>=3.8'
)
